use super::Keyfork;
use clap::{Args, Parser, Subcommand};

use keyfork_derive_openpgp::{
    openpgp::{
        armor::{Kind, Writer},
        packet::UserID,
        serialize::Marshal,
        types::KeyFlags,
    },
    XPrvKey,
};
use keyfork_derive_util::DerivationIndex;
use keyfork_derive_path_data::paths;
use keyforkd_client::Client;

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

#[derive(Subcommand, Clone, Debug)]
pub enum DeriveSubcommands {
    /// Derive an OpenPGP Transferable Secret Key (private key). The key is encoded using OpenPGP
    /// ASCII Armor, a format usable by most programs using OpenPGP.
    ///
    /// Certificates are created with a default expiration of one day, but may be configured to
    /// expire later using the `KEYFORK_OPENPGP_EXPIRE` environment variable using values such as
    /// "15d" (15 days), "1m" (one month), or "2y" (two years).
    ///
    /// It is recommended to use the default expiration of one day and to change the expiration
    /// using an external utility, to ensure the Certify key is usable.
    #[command(name = "openpgp")]
    OpenPGP(OpenPGP)
}

#[derive(Args, Clone, Debug)]
pub struct OpenPGP {
    /// Default User ID for the certificate, using the OpenPGP User ID format.
    user_id: String,
}

impl DeriveSubcommands {
    fn handle(&self, account: DerivationIndex) -> Result<()> {
        match self {
            DeriveSubcommands::OpenPGP(opgp) => opgp.handle(account),
        }
    }
}

impl OpenPGP {
    pub fn handle(&self, account: DerivationIndex) -> Result<()> {
        let path = paths::OPENPGP.clone().chain_push(account);
        // TODO: should this be customizable?
        let subkeys = vec![
            KeyFlags::empty().set_certification(),
            KeyFlags::empty().set_signing(),
            KeyFlags::empty()
                .set_transport_encryption()
                .set_storage_encryption(),
            KeyFlags::empty().set_authentication(),
        ];
        let xprv = Client::discover_socket()?.request_xprv::<XPrvKey>(&path)?;
        let default_userid = UserID::from(self.user_id.as_str());
        let cert = keyfork_derive_openpgp::derive(xprv, &subkeys, &default_userid)?;

        let mut w = Writer::new(std::io::stdout(), Kind::SecretKey)?;

        for packet in cert.into_packets2() {
            packet.serialize(&mut w)?;
        }

        w.finalize()?;
        Ok(())
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Derive {
    #[command(subcommand)]
    command: DeriveSubcommands,

    /// Account ID. Required for all derivations.
    ///
    /// An account ID may not be relevant for the derivation being performed, but the lack of an
    /// account ID can often come as a hindrance in the future. As such, it is always required. If
    /// the account ID is not relevant, it is assumed to be `0`.
    #[arg(long, global = true, default_value = "0")]
    account_id: u32,
}

impl Derive {
    pub fn handle(&self, _k: &Keyfork) -> Result<()> {
        let account = DerivationIndex::new(self.account_id, true)?;
        self.command.handle(account)
    }
}
